(* FAIRTHREADS *)

(* implantation naive -> optimisations possibles *)

(****************************************)
(* Declarations des types et exceptions *)
(****************************************)

exception No_value
exception FT_error of string
exception FT_stop

(* type d'identifiant pour les threads *)
type id = int

(* statuts possibles des threads *)
type status_t = | CONT (* l'execution du thread se poursuit dans l'instant courant *) 
      		| TERM (* l'execution du thread est totalement terminee *)
		| COOP (* l'execution du thread est terminee pour l'instant courant *)

(* base pour les threads et les schedulers *)
type base_t =
    {
      thread : Thread.t ; (* PThread correspondant *)
      cond : Condition.t; (* condition pour activation du thread *)
      mutex : Mutex.t;    (* mutex du thread *)
      mutable token : bool; (* token : vrai quand le thread a la main *)
    }

(* type des evenements *)
type 'a event_t = id * scheduler_t * 'a list ref
		     
(* type des schedulers *)
and scheduler_t =
    { 
      sbase : base_t ;

      mutable eoi : bool ; (* true quand fin d'instant *)
      mutable move : bool ; (* passe a true quand nouvel evenement *)
      mutable linked_thread : int list ; (* liste des threads lies *)
      mutable current : int list ; (* liste des threads pour la boucle courante *)
      mutable next : int list ; (* liste des threads pour la boucle suivante *)
      mutable to_start :  int list ; (* threads a demarrer au prochain instant *)
      mutable to_stop : int list ; (* threads a arreter au prochain instant *)
      mutable to_suspend : int list ; (* threads a suspendre au prochain instant *)
      mutable to_resume : int list ; (* threads a relancer au prochain instant *)
      mutable to_broadcast : id list ; (* evenements a generer au prochain instant *)
      mutable to_broadcast_value : id list ;  (* valeurs a generer au prochain instant *)
      mutable events : id list ; (* evenements dans l'instant courant *)
      mutable values : id list ; (* valurs dans l'instant courant *)

      mutable temps : int; (* compteur d'instants : pas reellement utilise *)
    }

(* type des threads *)
type thread_t =
    {
      tbase : base_t;

      mutable is_linked : bool ; (* thread li  un scheduler *)
      mutable sched : scheduler_t ; (* scheduler auquel le thread est li *)
      mutable status : status_t ; (* statut du thread *)
      mutable susp : bool (* vrai si le thread est suspendu *)
    }


(* condition utilisee pour le thread principal *)
let main_cond = Condition.create ()


(*************************************************************)
(* Modules de gestion des listes de threads et de schedulers *)
(*************************************************************)

(* on utilise un map pour stocker tous les fair threads et pour les
   scheduler *)
module M =  Map.Make(struct type t = int let compare = compare end)

(* stockage des threads *)
module ThList =
struct
  let mutex = Mutex.create ()
  let list = ref M.empty

  let add id thread =
    Mutex.lock mutex;
    list:= M.add id thread !list;
    Mutex.unlock mutex

  let remove id =
    Mutex.lock mutex;
    list:= M.remove id !list;
    Mutex.unlock mutex

  let find id =
    try
      M.find id !list
    with
	Not_found -> Printf.printf "Fthread error : Thread %d not found !\n" id;
	  flush stdout; raise Not_found

  let is_empty () =
    M.is_empty !list

  (* permet de recuperer le thread courant *)
  let this () =
    try
      M.find (Thread.id (Thread.self ())) !list
    with 
	Not_found -> Printf.printf "EXCEPTION : Erreur de this !\n"; flush stdout;
	  raise Not_found
end

(* stockage des schedulers *)
module SchList =
struct
  let mutex = Mutex.create ()

  let list = ref M.empty

  let add id sch =
    Mutex.lock mutex;
    list:= M.add id sch !list;
    Mutex.unlock mutex

  let remove id =
    Mutex.lock mutex;
    list:= M.remove id !list;
    Mutex.unlock mutex

  let find id =
    M.find id !list

  let is_empty () =
    M.is_empty !list

  (* permet de recuperer le scheduler courant *)
  let this () =
    try
      M.find (Thread.id (Thread.self ())) !list
    with
	Not_found -> Printf.printf "EXCEPTION : Erreur de this !\n";
	  raise Not_found
end



(****************************************************)
(* Fonctions de gestion des mutex et des conditions *)
(****************************************************)

let wait th =
  Condition.wait th.cond th.mutex

let lock th =
  Mutex.lock th.mutex

let unlock th =
  Mutex.unlock th.mutex

let signal th =
  Condition.signal th.cond

(* le scheduler donne la main a un thread *)
let give_token_to_thread thread =
  let th = thread.tbase in
  let sc = thread.sched.sbase in
    lock sc;
    lock th;
    th.token <- true;
    signal th;
    unlock th;
    while (not sc.token) do wait sc done;
    sc.token <- false;
    unlock sc
    
(* le thread rend la main au scheduler *)
let return_token_to_scheduler thread =
  let th = thread.tbase in
  let sc = thread.sched.sbase in
    lock th;
    lock sc;
    sc.token <- true;
    signal sc;
    unlock sc;
    while (not th.token) do wait th done;
    th.token <- false;
    unlock th;
    if thread.status = TERM then raise FT_stop
    
(* le thread rend la main au scheduler et est ajoute a la liste next *)
let return_token_and_add_to_next thread =
  let th = thread.tbase in
  let sch = thread.sched in
  let sc = sch.sbase in
    lock th;
    lock sc;
    sch.next <- (Thread.id th.thread) :: sch.next;
    sc.token <- true;
    signal sc;
    unlock sc;
    while (not th.token) do wait th done;
    th.token <- false;
    unlock th;
    if thread.status = TERM then raise FT_stop
    

(***********************************************************)
(* Fonctions d'affichage utilisees pour le debug du module *)
(***********************************************************)

(* fonction d'affichage d'un scheduler (DEBUG) *)  
let print_sch sch =
  Printf.printf "SCHEDULER %d\n" (Thread.id sch.sbase.thread);
  Printf.printf "| threads : ";
  List.iter (fun x -> Printf.printf "%d ; " x) sch.linked_thread;
  Printf.printf "\n\n";
  flush stdout

(* fonction d'affichage d'un thread (DEBUG) *) 
let print_th th =
  Printf.printf "THREAD %d\n" (Thread.id th.tbase.thread);
  Printf.printf "| status : ";
  match th.status with
      CONT -> Printf.printf "CONT\n\n"; flush stdout
    | COOP -> Printf.printf "COOP\n\n"; flush stdout
    | TERM -> Printf.printf "TERM\n\n"; flush stdout

(* fonction d'affichage d'un evenement *)
let print_event event =
  Printf.printf "EVENEMENT : ";
  Printf.printf "EVENT %d\n\n" event; flush stdout
    

(****************************)
(* fonctions de cooperation *)
(****************************)

(* cooperate : le thread a fini ses actions pour l'instant present *)
let cooperate () =
  let id = Thread.id (Thread.self()) in
  let th = ThList.find id in
    if th.is_linked then
      (
	lock th.tbase;
	th.status <- COOP;
	unlock th.tbase;
	return_token_to_scheduler th
      )

(* cooperate_n k : effectue l'action cooperate k fois *)
let rec cooperate_n k =
  if k > 0
  then
    (
      cooperate ();
      cooperate_n (k-1)
    )

(*******************************************)
(* fonctions pour la generation des events *)
(*******************************************)

let gen_event_id =
  let x = ref 0 in
  let f () = x := !x +1; !x in
  f

(* creation d'un evenement *)
let (create_event : scheduler_t -> 'a event_t) =
  fun s -> (gen_event_id (), s, ref [])

(* genere un evenement  *)
let (generate : 'a event_t -> unit) =
  fun (id,s,_) ->
    let thread = ThList.this() in
      if thread.is_linked then
	begin
	  let sched = thread.sched in
	    if sched = s then
	      begin
		let sc = sched.sbase in
		  lock sc;
		  (* signale la generation d'un nouvel event *)
		  sched.move <- true;  
		  (* ajout de l'event a la liste *)
		  sched.events <- id :: sched.events ; 
		  unlock sc
	      end
	    else
	      raise (FT_error "event used with incorrect scheduler")
	end

(* genere un evenement avec une valeur associee *)
(* let generate_value (id,s,v) value = *)
let (generate_value : 'a event_t -> 'a -> unit) =
  fun (id,s,v) value ->
  let thread = ThList.this() in
    if thread.is_linked then
      (
	let sched = thread.sched in
	  if s = sched then 
	    (
	      let sc = sched.sbase in
		lock sc;
		(* signale la generation d'un nouvel event *) 
		sched.move <- true; 
		(* ajout de l'event aux listes *) 
		sched.events <- id :: sched.events ; 
		sched.values <- id :: sched.values ;
		(* v := value ;*)
		v := [value];
		unlock sc
	    )
	  else
	    raise (FT_error "event used with incorrect scheduler")
      )
      
(* le thread courrant attend l'evenement event *)
let await (id,s,_) = 
  let thread = ThList.this () in
    if thread.is_linked then
      ( 
	let sched = thread.sched in
	  if s = sched then
	    (
	      let absent() =
		not (List.exists (fun x -> id = x) sched.events) in 
		(* tant que event est absent, le thread rend la main *) 
		while absent() do return_token_and_add_to_next thread done
	    )
	  else
	    raise (FT_error "event used with incorrect scheduler")
      )
      
(* attente durant k instants *)
let rec await_n (id,s,_) k =
  let thread = ThList.this () in 
    if thread.is_linked then
      (
	let sched = thread.sched in
	  if s = sched then
	    (
	      let absent() =
		not (List.exists (fun x -> id = x) sched.events) in
		(* boucle d'attente *)
	      let rec f n =
		if (absent()) && n > 0 
		then 
		  ( 
		    return_token_and_add_to_next thread;
		    if sched.eoi 
		    then
		      (
			return_token_and_add_to_next thread;
 			f (n-1)
		      ) 
		    else 
 		      f n 
		  )
	      in
		f k
	    )
	  else
	    raise (FT_error "event used with incorrect scheduler")
	      (* l'evenement utilise ne depend pas du scheduler du
		 thread courant *)
      )
     
(* recuperation de la premiere valeur associee a un event *)
(* let get_value (id,s,v) = *)
let (get_value : 'a event_t -> 'a) =
  fun (id,s,v) ->
  let thread = ThList.this () in
    if thread.is_linked then
      (
	let sched = thread.sched in
	  if s = sched then
	    (
	      let absent() =
		not (List.exists (function x -> id = x) sched.values) in
		(* tant que event est absent ou n'a pas de valeur, le thread
		   rend la main *)
		while absent()
		do
		  return_token_and_add_to_next thread
		done;
		try List.hd (!v) 
		with Failure "hd" -> raise No_value
	    )
	  else
	    raise (FT_error "event used with incorrect scheduler")
	      (* l'evenement utilise ne depend pas du scheduler du
		 thread courant *)
      )
    else
      raise (FT_error "error : an unliked thread can't wait for an event")
	(* si le thread n'est pas lie : erreur *)

(* let get_value_n (id,s,v) k = *)
let (get_value_n : 'a event_t -> int -> 'a) =
  fun (id,s,v) k ->
  let thread = ThList.this () in
    if thread.is_linked then
      (
	let sched = thread.sched in
	  if s = sched then
	    (
	      let absent() =
		not (List.exists (function (x) -> id = x) sched.values) in
		(* boucle d'attente *)
	      let rec f n =
		if n > 0 then
		  (
		    if absent () then
		      (
			return_token_and_add_to_next thread;
			if sched.eoi then
			  (
			    return_token_and_add_to_next thread;
			    f (n-1)
			  )
			else f n
		      )
		    else true
		      (* la valeur est presente *)
		  )
		else false 
		  (* pas de valeur apres le nombre d'instants specifie *)
	      in
		if f k then 
		  try List.hd (!v) 
		  with Failure "hd" -> raise No_value
		else raise No_value
		  (* renvoie une exception si apres l'attente aucune
		     valeur n'est disponible *)
	    )
	  else raise (FT_error "event used with incorrect scheduler")
	    (* l'evenement utilise ne depend pas du scheduler du
	       thread courant *)
      )
    else
      raise (FT_error "error : an unliked thread can't wait for an event")
	(* si le thread n'est pas lie, erreur *)

let broadcast : 'a event_t -> unit =
  fun (id, s, _) ->
    lock s.sbase;
    s.to_broadcast <- id :: s.to_broadcast;
    unlock s.sbase

let broadcast_value : 'a event_t -> 'a -> unit =
  fun (id, s, v) value ->
    lock s.sbase;
    v := [value];
    s.to_broadcast_value <- id :: s.to_broadcast_value;
    unlock s.sbase
      

(****************************)
(* fonctions link et unlink *)
(****************************)

(* detache le thread courrant de son scheduler *)
let unlink () =
  let thread = ThList.this() in
    if thread.is_linked then
      begin
	let th = thread.tbase in
	let sched = thread.sched in
	let sc = sched.sbase in
	let id = Thread.id th.thread in
	  lock th;
	  lock sc;
	  thread.is_linked <- false;
	  (* retire le thread de la liste du scheduler*)
	  sched.linked_thread <-
	    List.filter (fun elt -> elt != id) sched.linked_thread;
	  sc.token <- true;
	  signal sc;
	  unlock sc;
	  unlock th
      end

(* rattache le thread courrant a un scheduler *)
let link scheduler =
  let thread = ThList.this() in
    if thread.is_linked then unlink();
    let th = thread.tbase in
    let sc = scheduler.sbase in
      lock th;
      lock sc;
      thread.is_linked <- true;
      th.token <- false;
      thread.sched <- scheduler;
      (* ajout du thread au scheduler *)
      scheduler.to_start <- scheduler.to_start @ [Thread.id th.thread];
      unlock sc;
      (* attend que le scheduler lui donne la main *)
      while (not th.token) do wait th done;
      th.token <- false;
      unlock th
 
(*****************************************)
(*  Fonctions de controle des threads :  *)
(*     arret, suspension et reprise      *)
(*****************************************)
	
let thread_stop th =
  let sched = th.sched in
  let id = Thread.id th.tbase.thread in
    lock sched.sbase;
    sched.to_stop <- id :: sched.to_stop;
    unlock sched.sbase

let thread_suspend th =
  let sched = th.sched in
  let id = Thread.id th.tbase.thread in
    lock sched.sbase;
    sched.to_suspend <- id :: sched.to_suspend;
    unlock sched.sbase
      
let thread_resume th =
  let sched = th.sched in
  let id = Thread.id th.tbase.thread in
    lock sched.sbase;
    sched.to_resume <- id :: sched.to_resume;
    unlock sched.sbase

let join th = 
  Thread.join th.tbase.thread


(************************************************)
(*          Fonctions du scheduler :            *)
(* Gestion d'un instant et changement d'instant *)
(************************************************)

(* fonction de changement d'instant d'un scheduler *)
let next_instant () =
  let sch = SchList.this () in
    lock sch.sbase;
    (* ajout de to_start  linked_thread *)
    if sch.to_start != [] then sch.linked_thread <- sch.linked_thread @ sch.to_start;
    (* events remplac par to_broadcast *)
    sch.events <- sch.to_broadcast;
    sch.values <- sch.to_broadcast_value;
    (* on retire to_stop  linked_thread *)
    if sch.to_stop != [] then
      List.iter
	begin
	  fun elt ->
	    let thr = ThList.find elt in
	      thr.status <- TERM;
	      unlock sch.sbase;
	      give_token_to_thread thr;
	      lock sch.sbase
	end
	sch.to_stop;
    (* on retire les threads qui ont le statut TERM *)
    if sch.linked_thread != [] then
      sch.linked_thread <-
	List.filter
	begin
	  fun elt ->
	    let thr = ThList.find elt in
	    let res = thr.status != TERM in
	      if res then thr.status <- CONT
	      else ThList.remove elt;
	      res	
	end      
	sch.linked_thread;
    (* reactivation des threads de to_resume *)
    if sch.to_resume != [] then 
      List.iter 
	(fun x -> let t = ThList.find x in t.susp <- false)
	sch.to_resume;
    (* desactivation des threads de to_suspend *)
    if sch.to_suspend != [] then 
      List.iter 
	(fun x -> let t = ThList.find x in t.susp <- true) 
	sch.to_suspend;
    (* mets les statuts des threads  CONT *)
    List.iter 
      (fun x -> let t = ThList.find x in t.status <- CONT) 
      sch.linked_thread;
    (* vide les listes *)
    sch.to_start <- [];
    sch.to_broadcast <- [];
    sch.to_broadcast_value <- [];
    sch.to_stop <- [];
    sch.to_suspend <- [];
    sch.to_resume <- [];
    sch.eoi <- false;
    sch.temps <- sch.temps +1;
    (* preparation des listes pour la fonction instant *)
    sch.current <- List.filter
      (fun elt -> let th = ThList.find elt in not th.susp)
      sch.linked_thread;
    sch.next <- [];
    unlock sch.sbase


(* deroulement d'un instant *)
let instant () =
  let sch = SchList.this () in
  let continue = ref true in
    while !continue
    do
      lock sch.sbase;
      sch.move <- false;
      unlock sch.sbase;
      for i=0 to ((List.length sch.current)-1)
      do
	flush stdout;
	let th = ThList.find (List.nth sch.current i) in
	  if (th.status = CONT && not th.susp)
	  then give_token_to_thread th
      done;
      lock sch.sbase;
      if sch.eoi
      then
	continue := false
      else 
	begin
	  sch.current <- List.rev sch.next;
	  sch.next <- [];
	  if not sch.move then sch.eoi <- true;
	end;
      unlock sch.sbase
    done


let start_scheduler sched =
  let sc = sched.sbase in
    lock sc;
    sc.token <- true;
    signal sc;
    unlock sc


(* base d'un fairthread ou d'un scheduler *)
let create_base c m thr =
  {
    thread = thr;
    cond = c;
    mutex = m;
    token = false
  }


(* creation d'un scheduler *)
let create_scheduler () =
  let m = Mutex.create () in
  let c = Condition.create () in
  Mutex.lock m;
  let boucle _ = 
    begin
      Mutex.lock m; 
      let sc = SchList.this () in
      let base = sc.sbase in
      let id = Thread.id base.thread in
      signal base;
      (* attente du lancement initial *)
      while not base.token do wait base done;
      base.token <- false; 
      unlock base;
      (* boucle principale *)
       let continuer = ref true in  
	 while !continuer 
	do
	  instant ();
	  next_instant ();
	  (* si aucun thread  executer dans le scheduler -> arret *)
	  if (List.length sc.linked_thread) = 0 then continuer := false
	done;
	SchList.remove id;
	Condition.broadcast main_cond
    end 
  in
  let thr = Thread.create boucle 0 in
  let id = Thread.id thr in
  let base = create_base c m thr in
  let sch =
  {
    sbase = base;
    current = [];
    next = [];
    linked_thread = [];
    events = [];
    eoi = false;
    move = false;
    to_broadcast = [];
    to_broadcast_value = [];
    to_start = [];
    to_stop = [];
    to_suspend = [];
    to_resume = [];
    values = [];
    temps = 0
  } in 
  SchList.add id sch;
  wait base;
  unlock base;
  sch


(* fin normale d'un thread *)
let th_end thread =
  let th = thread.tbase in
  let sched = thread.sched in
  let sc = sched.sbase in
(*  let id = Thread.id th.thread in *)
(*  generate (TERM_EVT id); *)
  lock th;
  lock sc;
  thread.status <- TERM;
  sc.token <- true;
  signal sc;
  unlock sc;
  unlock th


(* creation d'un fair thread *)
let create_thread sch f arg =
  let c = Condition.create () in
  let m = Mutex.create () in
  Mutex.lock m;
  let exec _ =
    begin
      Mutex.lock m;
      let th = ThList.this () in
      let base = th.tbase in
      (* mise en attente avant lancement initial *)
      signal base;
      while not base.token do wait base done;
      base.token <- false;
      unlock base;
      (* fonction principale *)
      begin
	try
	  f arg
	with
	  | FT_stop -> ()
	  | x -> th_end th; ignore (raise x )
      end;
      (* phase terminale *)
      th_end th
    end
  in
  let thr = Thread.create exec 0 in
  let id = Thread.id thr in
  let base = create_base c m thr in
  let fthr =
  { tbase = base;
    sched = sch;
    status = CONT;
    susp = false;
    is_linked =  true;
  } in
  ThList.add id fthr;
  sch.to_start <- sch.to_start @ [id];
  wait base;
  unlock base;
  fthr


(* a mettre a la fin d'un programme *)
let exit () =
(*   Printf.printf "-> exit\n"; flush stdout; *)
  let m = Mutex.create () in
    Mutex.lock m;
    while (not (SchList.is_empty ()) )
    do
      Condition.wait main_cond m;
    done;
    Mutex.unlock m


